"""Mutation initialization.
"""
__version__ = "0.4.2"

__title__ = "mutatest"
__description__ = "Python mutation testing."
__url__ = "https://github.com/EvanKepner/mutatest"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Evan Kepner"

# matches my GitHub commit email, open issues on the repo for comments or questions
__email__ = "EvanKepner@users.noreply.github.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2018 Evan Kepner"
